<?php

namespace Drupal\memcache_storage;

use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Site\Settings;
use Psr\Log\LogLevel;

class DrupalMemcache extends DrupalMemcachedBase {

  protected $persistentConnection;

  protected $compressThreshold;

  protected $extension = 'Memcache';

  protected $compressionEnabled;

  /**
   * @var \Memcache.
   */
  protected $memcached;

  public function __construct(Settings $settings, array $servers, $cluster_name) {
    parent::__construct($settings, $servers, $cluster_name);

    $this->persistentConnection = $settings->get('memcache_storage.persistent_connection', FALSE);

    // For more information see http://www.php.net/manual/en/memcache.setcompressthreshold.php.
    $this->compressThreshold = $settings->get('memcache_storage.compress_threshold', ['threshold' => 20000, 'min_savings' => 0.2]);
    if (isset($this->compressThreshold['threshold']) && isset($this->compressThreshold['min_savings'])) {
      $this->memcached->setCompressThreshold($this->compressThreshold['threshold'], $this->compressThreshold['min_savings']);
    }

    $this->compressionEnabled = $settings->get('memcache_storage.compression_enabled');
  }

  public function isConnected() {
    $stats = $this->memcached->getStats();
    if (empty($stats) || empty($stats['uptime'])) {
      // TODO: move to shutdown.
      DrupalMemcachedUtils::log(LogLevel::ERROR, 'Could not connect to all servers from the cluster %cluster.', ['%cluster' => $this->clusterName]);
      return FALSE;
    }

    return TRUE;
  }

  public function addServer($host, $port) {
    // Add a new server to the memcached pool.
    $this->memcached->addServer($host, $port, $this->persistentConnection);
  }

  public function setMulti(array $items, $cache_bin = '') {

    foreach ($items as $item) {
      $memcached_key = $this->itemKey($item->cid, $cache_bin);

      $expiration = $item->expire;
      if ($item->expire == CacheBackendInterface::CACHE_PERMANENT) {
        $expiration = 0;
      }

      $compression = !empty($this->compressionEnabled) ? MEMCACHE_COMPRESSED : 0;
      $this->memcached->set($memcached_key, $item, $compression, $expiration);
    }
  }

  public function getMulti(array $keys, $cache_bin = '') {

    $memcached_keys = [];
    foreach ($keys as $key) {
      $memcached_key = $this->itemKey($key, $cache_bin);
      $memcached_keys[$memcached_key] = $key;
    }

    $compression = !empty($this->compressionEnabled) ? MEMCACHE_COMPRESSED : 0;
    $result = $this->memcached->get(array_keys($memcached_keys), $compression);

    $cache = [];
    foreach ($result as $memcached_key => $value) {
      $normal_key = $memcached_keys[$memcached_key];
      $cache[$normal_key] = $value;
    }

    return $cache;
  }

  public function deleteMulti(array $keys, $cache_bin = '') {
    foreach ($keys as $key) {
      $memcached_key = $this->itemKey($key, $cache_bin);
      $this->memcached->delete($memcached_key);
    }
  }

}
