<?php

namespace Drupal\memcache_storage;

use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Site\Settings;
use Psr\Log\LogLevel;

class DrupalMemcached extends DrupalMemcachedBase {

  protected $options;

  protected $saslAuth;

  protected $extension = 'Memcached';

  protected $optionsDefault = [
    \Memcached::OPT_COMPRESSION => FALSE,
    \Memcached::OPT_DISTRIBUTION => \Memcached::DISTRIBUTION_CONSISTENT,
  ];

  /**
   * @var \Memcached.
   */
  protected $memcached;

  public function __construct(Settings $settings, array $servers, $cluster_name) {
    parent::__construct($settings, $servers, $cluster_name);

    // For more info about memcached constants see http://www.php.net/manual/en/memcached.constants.php.
    $this->options = $settings->get('memcache_storage.memcached_options', []);
    $this->options += $this->optionsDefault;

    // Add SASL support.
    $this->saslAuth = $settings->get('memcache_storage.sasl_auth', []);
    if (!empty($sasl_auth['user']) && !empty($sasl_auth['password'])) {
      $this->memcached->setSaslAuthData($sasl_auth['user'], $sasl_auth['password']);

      // SASL auth works only with binary protocol.
      $this->options[\Memcached::OPT_BINARY_PROTOCOL] = TRUE;
    }

    // Set memcached options.
    // See http://www.php.net/manual/en/memcached.setoption.php.
    foreach ($this->options as $key => $value) {
      $this->memcached->setOption($key, $value);
    }
  }

  public function isConnected() {
    $stats = $this->memcached->getStats();
    if (empty($stats)) {
      return FALSE;
    }

    $failed_connections = [];
    foreach ($stats as $server => $server_stats) {
      if (empty($server_stats['uptime'])) {
        $failed_connections[] = $server;
        // TODO: move to shutdown.
        DrupalMemcachedUtils::log(LogLevel::WARNING, 'Could not connect to the server %server.', ['%server' => $server]);
      }
    }

    // If memcached is unable to connect to all servers it means that
    // we have no connections at all, and could not start memcached connection.
    if (sizeof($failed_connections) == sizeof($stats)) {
      // TODO: move to shutdown.
      DrupalMemcachedUtils::log(LogLevel::ERROR, 'Could not connect to all servers from the cluster %cluster.', ['%cluster' => $this->clusterName]);
      return FALSE;
    }

    return TRUE;
  }

  public function addServer($host, $port) {
    // Add a new server to the memcached pool.
    $this->memcached->addServer($host, $port);
  }

  public function setMulti(array $items, $cache_bin = '') {

    $item_groups = [];
    foreach ($items as $item) {
      $memcached_key = $this->itemKey($item->cid, $cache_bin);

      $expiration = $item->expire;
      if ($item->expire == CacheBackendInterface::CACHE_PERMANENT) {
        $expiration = 0;
      }

      $item_groups[$expiration][$memcached_key] = $item;
    }

    foreach ($item_groups as $expiration => $cache_items) {
      $this->memcached->setMulti($cache_items, $expiration);
    }
  }

  public function getMulti(array $keys, $cache_bin = '') {

    $memcached_keys = [];
    foreach ($keys as $key) {
      $memcached_key = $this->itemKey($key, $cache_bin);
      $memcached_keys[$memcached_key] = $key;
    }

    $result = $this->memcached->getMulti(array_keys($memcached_keys));

    $cache = [];
    foreach ($result as $memcached_key => $value) {
      $normal_key = $memcached_keys[$memcached_key];
      $cache[$normal_key] = $value;
    }

    return $cache;
  }

  public function deleteMulti(array $keys, $cache_bin = '') {
    $memcached_keys = [];
    foreach ($keys as $key) {
      $memcached_keys[] = $this->itemKey($key, $cache_bin);
    }
    $this->memcached->deleteMulti($memcached_keys);
  }

}
