<?php

namespace Drupal\memcache_storage;

use Drupal\Core\Site\Settings;

abstract class DrupalMemcachedBase implements DrupalMemcachedInterface {

  /**
   * @var \Memcached or \Memcache object
   */
  protected $memcached;

  protected $settings;

  protected $keyPrefix;

  protected $hashAlgorithm;

  protected $clusterName;

  protected $extension;

  protected $bin_indexes;

  public function __construct(Settings $settings, array $servers, $cluster_name) {
    $this->memcached = new $this->extension();
    $this->settings = $settings;
    $this->clusterName = $cluster_name;
    $this->keyPrefix = $settings->get('memcache_storage.key_prefix');
    $this->hashAlgorithm = $settings->get('memcache_storage.key_hash_algorytm', 'sha1');

    foreach ($servers as $server) {
      list($host, $port) = DrupalMemcachedUtils::parseServerInfo($server);
      $this->addServer($host, $port);
    }
  }

  public function set($key, $value, $expiration = 0, $cache_bin = '') {
    $item = new \stdClass();
    $item->cid = $key;
    $item->expire = $expiration;
    $item->data = $value;
    $this->setMulti([$item], $cache_bin);
  }

  public function get($key, $cache_bin = '') {
    $cache = $this->getMulti([$key], $cache_bin = '');
    return reset($cache);
  }

  public function delete($key, $cache_bin = '') {
    $this->deleteMulti([$key], $cache_bin);
  }

  public function flush($cache_bin = '') {
    $this->increaseBinIndex($cache_bin);
  }

  final public function itemKey($key, $cache_bin) {

    // Build unique cache key.
    $cache_key  = !empty($this->keyPrefix) ? $this->keyPrefix . '-' : '';
    $cache_key .= $cache_bin ? $cache_bin . $this->getBinIndex($cache_bin) . '-' : '';
    $cache_key .= $key;
    $cache_key = urlencode($cache_key);

    // Memcache only supports key length up to 250 bytes. If we have generated
    // a longer key, hash it with md5 which will shrink the key down to 32 bytes
    // while still keeping it unique.
    if (strlen($cache_key) > 250) {
      $cache_key = urlencode(hash($this->hashAlgorithm, $cache_key));
    }

    return $cache_key;
  }

  /**
   * Increase cache bin index.
   * This operation changes all memcache keys in the specified cache bin,
   * so we fake the cache flush operation.
   */
  final protected function increaseBinIndex($cache_bin) {

    // We force reload the data from memcached to avoid race conditions with
    // other threads who could already increase bin index for the cache bin.
    $cache = $this->get('memcache_storage_bin_indexes');

    if (empty($cache->data)) {
      $this->bin_indexes[$cache_bin] = 1;
    }
    else {
      $this->bin_indexes = $cache->data;
      $current_index = !empty($cache->data[$cache_bin]) ? $cache->data[$cache_bin] : 0;
      $this->bin_indexes[$cache_bin] = ++$current_index;
    }

    $this->set('memcache_storage_bin_indexes', $this->bin_indexes, 0);
  }

  /**
   * Load cache bin index.
   * This index is part of memcache key and changes if cache bin should be cleared.
   */
  final protected function getBinIndex($cache_bin) {

    if (!isset($this->bin_indexes)) {
      $this->bin_indexes = [];

      // An attempt to get bin indexes from cache.
      $cache = $this->get('memcache_storage_bin_indexes');
      if (!empty($cache->data)) {
        $this->bin_indexes = $cache->data;
      }
    }

    if (empty($this->bin_indexes[$cache_bin])) {
      $this->bin_indexes[$cache_bin] = 1; // Initial index value.
      $this->set('memcache_storage_bin_indexes', $this->bin_indexes, 0);
    }

    return $this->bin_indexes[$cache_bin];
  }

}
